part of disposable;

/// Represents a disposable resource that only disposes its underlying
/// disposable resource when all [getDisposable] dependent disposable
/// objects have been disposed.
class RefCountDisposable implements Disposable {
  bool _isPrimaryDisposed = false;
  Disposable _disposable;
  int _count;

  /// Initializes a new instance of the [RefCountDisposable] class
  /// with the specified disposable.
  RefCountDisposable(Disposable disposable) {
    checkNotNull(disposable);

    _disposable = disposable;
    _isPrimaryDisposed = false;
    _count = 0;
  }

  @override
  bool get isDisposed => _disposable == null;

  /// Returns a dependent disposable that when disposed decreases the
  /// refcount on the underlying disposable.
  Disposable getDisposable() {
    if (_disposable == null) {
      return new Disposable();
    } else {
      _count++;
      return new _InnerDisposable(this);
    }
  }

  void _release() {
    var disposable = null;

    if (_disposable != null) {
      _count--;

      assert(_count >= 0);

      if (_isPrimaryDisposed) {
        if (_count == 0) {
          disposable = _disposable;
          _disposable = null;
        }
      }
    }

    if (disposable != null) disposable.dispose();
  }

  @override
  void dispose() {
    var disposable = null;

    if (_disposable != null) {
      if (!_isPrimaryDisposed) {
        _isPrimaryDisposed = true;

        if (_count == 0) {
          disposable = _disposable;
          _disposable = null;
        }
      }
    }

    if (disposable != null) disposable.dispose();
  }
}

class _InnerDisposable implements Disposable {
  RefCountDisposable _parent;

  _InnerDisposable(this._parent);

  @override
  bool get isDisposed => _parent == null;

  @override
  void dispose() {
    if (_parent != null) {
      _parent._release();
      _parent = null;
    }
  }
}
