part of disposable.async;

/// Provides a method of encapsulating a series of actions which ensure
/// proper disposal of the disposable instance prior to exiting but following
/// the completion of a Future value which is expected to be emitted from
/// the contained code block.
Future usingAsync(
    Disposable disposable, dynamic usingFunction(Disposable disposable)) {
  if (disposable.isDisposed) {
    return new Future.value(null);
  }
  dynamic result;
  dynamic error;
  if (disposable is AsyncDisposable) {
    Future future = disposable.future;
    try {
      result = usingFunction(disposable);
    } catch (ex) {
      _tryDisposeError(disposable, ex);
      return future;
    }

    if (result == null) {
      _tryDispose(disposable);
    } else if (result is Future) {
      result.then((value) {
        _tryDisposeWith(disposable, value);
      }, onError: (error) {
        _tryDisposeError(disposable, error);
      });
    } else {
      _tryDisposeWith(disposable, result);
    }
    return future;
  } else {
    Completer completer = new Completer();
    try {
      result = usingFunction(disposable);
    } catch (ex) {
      error = ex;
    }
    try {
      disposable.dispose();
    } catch (ex) {
      error = ex;
    }
    if (error != null) {
      completer.completeError(error);
    } else if (result == null) {
      completer.complete();
    } else {
      completer.complete(result);
    }
    return completer.future;
  }
  return new Future.value(null);
}

void _tryDispose(Disposable disposable) {
  try {
    disposable.dispose();
  } catch (ex) {
    if (disposable is AsyncDisposable) {
      disposable.disposeError(ex);
    }
  }
}

void _tryDisposeWith(AsyncDisposable disposable, dynamic value) {
  try {
    disposable.disposeWith(value);
  } catch (ex) {
    disposable.disposeError(ex);
  }
}

void _tryDisposeError(AsyncDisposable disposable, dynamic error) {
  try {
    disposable.disposeError(error);
  } catch (ex) {
    disposable.disposeError(ex);
  }
}
