library disposable_test.composite_disposable;

import 'package:disposable/disposable.dart';
import 'package:test/test.dart';

main() {
  group('-composite_disposable- should', () {
    test('throw an ArgumentError if adding null to composite', () {
      CompositeDisposable compositeDisposable = new CompositeDisposable();

      expect(() => compositeDisposable.add(null), throwsArgumentError);
    });

    test('throw an ArgumentError if removing null from composite', () {
      CompositeDisposable compositeDisposable = new CompositeDisposable();

      expect(() => compositeDisposable.remove(null), throwsArgumentError);
    });

    test('returns false if removing a non-existent disposable from composite', () {
      CompositeDisposable compositeDisposable = new CompositeDisposable();
      Disposable disposable = new Disposable();

      expect(compositeDisposable.remove(disposable), isFalse);
    });

    test('throw an ArgumentError if checking for contains null from composite', () {
      CompositeDisposable compositeDisposable = new CompositeDisposable();

      expect(() => compositeDisposable.contains(null), throwsArgumentError);
    });

    test('trigger mulitiple callbacks from CompositeDisposable', () {
      // Arrange
      int disposedCount = 0;
      var callback = () {
        disposedCount++;
      };
      Disposable disposable1 = new Disposable.create(callback);
      Disposable disposable2 = new Disposable.create(callback);
      Disposable disposable3 = new Disposable.create(callback);

      final compositeDisposable =
          new CompositeDisposable.from([disposable1, disposable2, disposable3]);
      expect(compositeDisposable.length, 3);

      // Act
      compositeDisposable.dispose();

      // Assert
      expect(disposedCount, 3);
    });

    test('add multiple disposables at once', () {
      // Arrange
      int disposedCount = 0;
      var callback = () {
        disposedCount++;
      };
      Disposable disposable1 = new Disposable.create(callback);
      Disposable disposable2 = new Disposable.create(callback);
      Disposable disposable3 = new Disposable.create(callback);

      final compositeDisposable = new CompositeDisposable();
      compositeDisposable.addAll([disposable1, disposable2, disposable3]);
      expect(compositeDisposable.length, 3);

      compositeDisposable.dispose();
      expect(disposedCount, 3);
    });

    test('dispose items aadded prior to composite disposal with the container',
        () {
      // Arrange
      int disposedCount = 0;
      var callback = () {
        disposedCount++;
      };
      Disposable disposable1 = new Disposable.create(callback);
      Disposable disposable2 = new Disposable.create(callback);

      CompositeDisposable compositeDisposable = new CompositeDisposable();
      compositeDisposable.add(disposable1);
      compositeDisposable.add(disposable2);
      expect(disposedCount, 0);

      // Act
      compositeDisposable.dispose();

      // Assert
      expect(disposedCount, 2);
    });

    test(
        'dispose all other items even if a callback throws an exception on composite dispose',
        () {
      // Arrange
      bool caughtException = false;
      int disposedCount = 0;
      var callback = () {
        disposedCount++;
      };
      Disposable disposable1 = new Disposable.create(callback);
      Disposable disposable2 = new Disposable.create(callback);
      Disposable disposableThrower =
          new Disposable.create(() => throw 'error');

      CompositeDisposable compositeDisposable = new CompositeDisposable();
      compositeDisposable.add(disposable1);
      compositeDisposable.add(disposableThrower);
      compositeDisposable.add(disposable2);
      expect(disposedCount, 0);

      // Act
      try {
        compositeDisposable.dispose();
      } catch (ex) {
        caughtException = true;
      }
      // Assert
      expect(disposedCount, 2);
      expect(caughtException, true);
    });

    test(
        'keep failed disposable in composite if a callback throws an exception on composite clear',
        () {
      // Arrange
      int disposedCount = 0;
      var callback = () {
        disposedCount++;
      };
      Disposable disposable1 = new Disposable.create(callback);
      Disposable disposable2 = new Disposable.create(callback);
      Disposable disposableThrower =
          new Disposable.create(() => throw 'error');

      CompositeDisposable compositeDisposable = new CompositeDisposable();
      compositeDisposable.add(disposable1);
      compositeDisposable.add(disposableThrower);
      compositeDisposable.add(disposable2);
      expect(disposedCount, 0);

      expect(() => compositeDisposable.clear(), throws);
      expect(disposedCount, 2);
    });

    test(
        'keep reference to disposable in error if a callback throws an exception on composite clear',
        () {
      // Arrange
      var error = null;
      int disposedCount = 0;
      var callback = () {
        disposedCount++;
      };
      Disposable disposable1 = new Disposable.create(callback);
      Disposable disposable2 = new Disposable.create(callback);
      Disposable disposableThrower =
          new Disposable.create(() => throw 'error');

      CompositeDisposable compositeDisposable = new CompositeDisposable();
      compositeDisposable.add(disposable1);
      compositeDisposable.add(disposableThrower);
      compositeDisposable.add(disposable2);
      expect(disposedCount, 0);

      // Act
      try {
        compositeDisposable.clear();
      } catch (ex) {
        error = ex as List;
      }
      // Assert
      expect(error.length, 1);
    });

    test(
        'immediately dispose items added to a CompositeDisposable which was already disposed',
        () {
      // Arrange
      int disposedCount = 0;
      var callback = () {
        disposedCount++;
      };
      Disposable disposable = new Disposable.create(callback);

      CompositeDisposable compositeDisposable = new CompositeDisposable();
      compositeDisposable.dispose();

      expect(disposedCount, 0, reason: 'should not have been disposed yet');
      expect(compositeDisposable.length, equals(0));

      // Act
      compositeDisposable
          .add(new Disposable.create(() => disposedCount++));

      // Assert
      expect(disposedCount, 1,
          reason:
              'should have been disposed without calling dispose on the composite');
      expect(compositeDisposable.length, equals(0));
    });

    test(
        'dispose and remove item from CopositeDisposable when remove is called',
        () {
      // Arrange
      int disposedCount = 0;
      var callback = () {
        disposedCount++;
      };
      Disposable disposable = new Disposable.create(callback);

      CompositeDisposable compositeDisposable =
          new CompositeDisposable.from([disposable]);
      expect(disposedCount, 0, reason: 'should not have been disposed yet');

      // Act
      bool removed = compositeDisposable.remove(disposable);

      // Assert
      expect(removed, isTrue);
      expect(disposedCount, 1, reason: 'should not have been disposed yet');
    });

    test(
        'dispose and remove all items from CompositeDisposable without triggering a dispose on the composite itself',
        () {
      // Arrange
      int disposedCount = 0;
      var callback = () {
        disposedCount++;
      };
      Disposable disposable1 = new Disposable.create(callback);
      Disposable disposable2 = new Disposable.create(callback);

      CompositeDisposable compositeDisposable =
          new CompositeDisposable.from([disposable1]);

      expect(disposedCount, 0,
          reason: 'first should not have been disposed yet');

      // Act
      compositeDisposable.clear();
      expect(disposedCount, 1,
          reason: 'second should not have been disposed yet');
      compositeDisposable.add(disposable2);

      // Assert
      expect(disposedCount, 1, reason: 'second should not have been disposed');
    });
  });
}
