library disposable_test.using_async;

import 'dart:async';

import 'package:disposable/async.dart';
import 'package:disposable/disposable.dart';
import 'package:test/test.dart';

class Thing implements AsyncDisposable {
  final BooleanAsyncDisposable _disposable;
  bool _isDisposed = false;

  bool get isDisposed => _isDisposed;
  Future get future => _disposable.future;

  Thing._(BooleanAsyncDisposable this._disposable);

  factory Thing() {
    Thing thing;
    BooleanAsyncDisposable disposable =
        new BooleanAsyncDisposable.create(() {});
    thing = new Thing._(disposable);
    return thing;
  }

  void _dispose() {
    _isDisposed = true;
  }

  void dispose() {
    _dispose();
    _disposable.dispose();
  }

  void disposeWith(dynamic value) {
    _dispose();
    _disposable.disposeWith(value);
  }

  void disposeError(dynamic error) {
    _dispose();
    _disposable.disposeError(error);
  }
}

main() {
  group('-using_async- should', () {
    test('dispose of Disposable', () {
      // Arrange
      bool disposed = false;
      Disposable disposable =
          new Disposable.create(() => disposed = true);

      // Act
      usingAsync(disposable, (Disposable disp) {
        expect(disposed, false);
      }).then((value) {
        expect(disposed, true);
      }).then(expectAsync((_) {}));
    });

    test('still trigger dispose if wrapped block throws exception', () {
      // Arrange
      bool disposed = false;
      Disposable disposable =
          new Disposable.create(() => disposed = true);
      // Act
      usingAsync(disposable, (Disposable disposable) {
        throw 'error';
      }).then((value) {}, onError: (error) {}).then(expectAsync((_) {}));

      // Assert
      expect(disposed, true,
          reason: 'still trigger dispose if wrapped block throws exception');
    });

    test('rethrow caught exception after disposing', () {
      // Arrange
      bool disposed = false;
      bool caughtError = false;
      var error = null;
      Disposable disposable =
          new Disposable.create(() => disposed = true);

      // Act
      usingAsync(disposable, (Disposable disposable) {
        throw 'error';
      }).then((value) {}, onError: (error) {
        // Assert
        expect(disposed, true);
        expect(error, 'error');
      });
    });

    test(
        'trigger Future.completeError if disposable is IAsyncDisposable and exception is thrown in wrapped block',
        () {
      // Arrange
      bool disposed = false;
      bool futureCalled = false;
      var error = null;
      AsyncDisposable disposable =
          new BooleanAsyncDisposable.create(() => disposed = true);

      var future = disposable.future
          .then((_) => futureCalled = true)
          .catchError((ex) => error = ex)
          .then(expectAsync((_) {
        expect(futureCalled, false);
        expect(error, 'error',
            reason:
                'if the future error handler was triggered this will not be null');
      }));

      // Act
      try {
        using(disposable, (AsyncDisposable disposable) {
          throw 'error';
        });
      } catch (ex) {}

      expect(disposed, true);
    });

    test('dispose of AsyncDisposable implementation with future return value',
        () {
      // Arrange
      Thing thing = new Thing();

      expect(thing.isDisposed, false);
      // Act
      usingAsync(thing, (Thing thing) {
        expect(thing.isDisposed, false);
      }).then((value) {
        expect(thing.isDisposed, true);
      }).then(expectAsync((_) {}));
    });

    test('return future<null> if disposable was already disposed', () {
      // Arrange
      bool disposed = false;
      AsyncDisposable disposable =
          new BooleanAsyncDisposable.create(() => disposed = true);
      disposable.dispose();

      // Act
      usingAsync(disposable, (AsyncDisposable disposable) {
        // Already disposed
        expect(false, true, reason: 'should not get here');
      }).then((value) {
        // Assert
        expect(value, null);
      }).then(expectAsync((_) {}));
    });

    test('allow nothing response from wrapped code without blocking', () {
      // Arrange
      bool disposed = false;
      AsyncDisposable disposable =
          new BooleanAsyncDisposable.create(() => disposed = true);

      // Act
      usingAsync(disposable, (AsyncDisposable disposable) {
        expect(disposed, false,
            reason: 'should not have disposed inside the wrapped');
        // nothing...
      }).then((value) {
        // Assert
        expect(disposed, true,
            reason:
                'should have disposed prior to completion without blocking');
      }, onError: (error) {
        expect(false, true,
            reason: 'should not complete with an exception: ${error}');
      }).then(expectAsync((_) {}));
    });

    test('complete with result if Future is returned', () {
      // Arrange
      bool disposed = false;
      AsyncDisposable disposable =
          new BooleanAsyncDisposable.create(() => disposed = true);
      Completer externalCompleter = new Completer();

      // Act
      usingAsync(disposable, (AsyncDisposable disposable) {
        return externalCompleter.future;
      }).then((value) {
        expect(disposed, true);
        expect(value, 'stuff');
      }, onError: (error) {
        // Assert
        expect(false, true,
            reason: 'should not complete with an exception: ${error}');
      }).then(expectAsync((_) {}));

      externalCompleter.complete('stuff');
    });

    /* ** Completer is blocked waiting for dispose
    test('complete with result if Future is completed inline', () {
      print('complete with result if Future is completed inline');
      // Arrange
      bool disposed = false;
      Completer externalCompleter = new Completer();
      IAsyncDisposable disposable = new AsyncDisposable(() {
        disposed = true;
        externalCompleter.complete('stuff');
      });

      // Act
      usingAsync(disposable, (IAsyncDisposable disposable) {
        return externalCompleter.future;
      })
      .then((value) {
        expect(disposed, true);
        expect(value, 'stuff');
      }, onError: (error) {
        // Assert
        expect(false, true, reason: 'should not complete with an exception: ${error}');
      })
      .then(expectAsync1((_) { }));
    });*/

    test('complete with result if Future is returned', () {
      // Arrange
      bool disposed = false;
      AsyncDisposable disposable =
          new BooleanAsyncDisposable.create(() => disposed = true);
      Completer externalCompleter = new Completer();

      // Act
      usingAsync(disposable, (AsyncDisposable disposable) {
        return 'stuff';
      }).then((value) {
        expect(value, 'stuff');
      }, onError: (error) {
        // Assert
        expect(false, true,
            reason: 'should not complete with an exception: ${error}');
      }).then(expectAsync((_) {}));
    });

    test(
        'complete with onError from Future if exception is thrown in wrapped code',
        () {
      // Arrange
      bool disposed = false;
      AsyncDisposable disposable =
          new BooleanAsyncDisposable.create(() => disposed = true);

      // Act
      usingAsync(disposable, (disposable) {
        expect(disposed, false,
            reason: 'should not have disposed inside the wrapped');
        throw 'error';
      }).then((value) {
        expect(false, true,
            reason: 'should complete with an exception if wrapped throws');
      }, onError: (error) {
        // Assert
        expect(error, 'error');
        expect(disposed, true,
            reason:
                'should have disposed prior to completion with error: ${error}');
      }).then(expectAsync((_) {}));
    });

    test(
        'complete with onError from Future if exception is thrown in wrapped code and dispose',
        () {
      // Arrange
      bool disposed = false;
      AsyncDisposable disposable =
          new BooleanAsyncDisposable.create(() => throw 'dispose error');

      // Act
      usingAsync(disposable, (disposable) {
        expect(disposed, false,
            reason: 'should not have disposed inside the wrapped');
        throw 'error';
      }).then((value) {
        expect(false, true,
            reason:
                'should complete with an exception if both wrapped and dispose throw');
      }, onError: (error) {
        // Assert
        expect(error, 'dispose error');
      }).then(expectAsync((_) {}));
    });

    test(
        'complete with onError from Future if exception is thrown in dispose but not in wrapped code',
        () {
      // Arrange
      bool disposed = false;
      AsyncDisposable disposable =
          new BooleanAsyncDisposable.create(() => throw 'dispose error');

      // Act
      usingAsync(disposable, (disposable) {
        expect(disposed, false,
            reason: 'should not have disposed inside the wrapped');
        // nothing...
      }).then((value) {
        expect(false, true, reason: 'should complete with an exception');
      }, onError: (error) {
        // Assert
        expect(error, 'dispose error');
      }).then(expectAsync((_) {}));
    });
  });
}
